sinon = require('sinon')
chai = require('chai')
should = chai.should()
expect = chai.expect
modulePath = "../../../../app/js/HistoryManager.js"
SandboxedModule = require('sandboxed-module')

describe "HistoryManager", ->
	beforeEach ->
		@HistoryManager = SandboxedModule.require modulePath, requires:
			"./UpdateCompressor": @UpdateCompressor = {}
			"./MongoManager" : @MongoManager = {}
			"logger-sharelatex": { log: sinon.stub() }
		@doc_id = "doc-id-123"
		@callback = sinon.stub()

	describe "when there are no raw ops", ->
		beforeEach ->
			@MongoManager.popLastCompressedUpdate = sinon.stub()
			@MongoManager.insertCompressedUpdates = sinon.stub()
			@HistoryManager.compressAndSaveRawUpdates @doc_id, [], @callback

		it "should not need to access the database", ->
			@MongoManager.popLastCompressedUpdate.called.should.equal false
			@MongoManager.insertCompressedUpdates.called.should.equal false

		it "should call the callback", ->
			@callback.called.should.equal true

	describe "when there is no compressed history to begin with", ->
		beforeEach ->
			@rawUpdates = [{ v: 12, op: "mock-op-12" }, { v: 13, op: "mock-op-13" }]
			@compressedUpdates = { v: 13, op: "compressed-op-12" }

			@MongoManager.popLastCompressedUpdate = sinon.stub().callsArgWith(1, null, null)
			@MongoManager.insertCompressedUpdates = sinon.stub().callsArg(2)
			@UpdateCompressor.compressRawUpdates = sinon.stub().returns(@compressedUpdates)
			@HistoryManager.compressAndSaveRawUpdates @doc_id, @rawUpdates, @callback

		it "should try to pop the last compressed op", ->
			@MongoManager.popLastCompressedUpdate
				.calledWith(@doc_id)
				.should.equal true
		
		it "should compress the raw ops", ->
			@UpdateCompressor.compressRawUpdates
				.calledWith(null, @rawUpdates)
				.should.equal true
		
		it "should save the compressed ops", ->
			@MongoManager.insertCompressedUpdates
				.calledWith(@doc_id, @compressedUpdates)
				.should.equal true

		it "should call the callback", ->
			@callback.called.should.equal true

	describe "when the raw ops need appending to existing history", ->
		beforeEach ->
			@lastCompressedUpdate = { v: 11, op: "compressed-op-11" }
			@compressedUpdates = { v: 13, op: "compressed-op-12" }

			@MongoManager.popLastCompressedUpdate = sinon.stub().callsArgWith(1, null, @lastCompressedUpdate)
			@MongoManager.insertCompressedUpdates = sinon.stub().callsArg(2)
			@UpdateCompressor.compressRawUpdates = sinon.stub().returns(@compressedUpdates)

		describe "when the raw ops start where the existing history ends", ->
			beforeEach ->
				@rawUpdates = [{ v: 12, op: "mock-op-12" }, { v: 13, op: "mock-op-13" }]
				@HistoryManager.compressAndSaveRawUpdates @doc_id, @rawUpdates, @callback

			it "should try to pop the last compressed op", ->
				@MongoManager.popLastCompressedUpdate
					.calledWith(@doc_id)
					.should.equal true
			
			it "should compress the last compressed op and the raw ops", ->
				@UpdateCompressor.compressRawUpdates
					.calledWith(@lastCompressedUpdate, @rawUpdates)
					.should.equal true
			
			it "should save the compressed ops", ->
				@MongoManager.insertCompressedUpdates
					.calledWith(@doc_id, @compressedUpdates)
					.should.equal true

			it "should call the callback", ->
				@callback.called.should.equal true

		describe "when some raw ops are passed that have already been compressed", ->
			beforeEach ->
				@rawUpdates = [{ v: 10, op: "mock-op-10" }, { v: 11, op: "mock-op-11"}, { v: 12, op: "mock-op-12" }, { v: 13, op: "mock-op-13" }]

				@HistoryManager.compressAndSaveRawUpdates @doc_id, @rawUpdates, @callback

			it "should only compress the more recent raw ops", ->
				@UpdateCompressor.compressRawUpdates
					.calledWith(@lastCompressedUpdate, @rawUpdates.slice(-2))
					.should.equal true

		describe "when the raw ops do not follow from the last compressed op version", ->
			beforeEach ->
				@rawUpdates = [{ v: 13, op: "mock-op-13" }]
				@HistoryManager.compressAndSaveRawUpdates @doc_id, @rawUpdates, @callback

			it "should call the callback with an error", ->
				@callback
					.calledWith(new Error("Tried to apply raw op at version 13 to last compressed update with version 11"))
					.should.equal true


